from os import mkdir
from os.path import expanduser, join
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler

from cml.shared.errors import (
    NoModelReconstructedError,
    NotReducableLearnblockWarning
)
from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain.deconstruction import TS_QUEUE


class KnowledgeSearchUsecase(Usecase):
    def process(self, request):
        source, constr, selectr, reconstr, deconstr = self._init_loggers(
            request)
        reconstr.category = constr.construction_type

        id_generator = deconstr.knowledge_database.generate_free_ids()
        reconstr.free_id = id_generator
        deconstr.free_id = id_generator
        deconstr.reconstructor = reconstr
        self._run(source, constr, selectr, reconstr, deconstr)
        return deconstr.knowledge_database

    def _run(self, source, constructor, selector, reconstructor, deconstructor):
        tier = 1

        deconstructor.logger.protocol("{:#^100}".format("TIER 1"))
        for learnblock in source.learnblocks:
            for complete_learnblock in constructor.construct(learnblock):
                try:
                    reduced = selector.select(complete_learnblock)
                    pragmatic = reconstructor.reconstruct(tier, reduced)
                    deconstructor.deconstruct(tier, pragmatic, reduced)

                except NotReducableLearnblockWarning as error:
                    continue

                except NoModelReconstructedError as error:
                    continue

        while not TS_QUEUE.empty():
            new_tier, learnblock = TS_QUEUE.get()
            if new_tier > tier:
                banner = "TIER {}".format(new_tier)
                deconstructor.logger.protocol("{:#^100}".format(banner))
                tier = new_tier

            for complete_learnblock in constructor.construct(learnblock):
                try:
                    pragmatic = reconstructor.reconstruct(tier, complete_learnblock)
                    deconstructor.deconstruct(tier, pragmatic, complete_learnblock)

                except NotReducableLearnblockWarning as error:
                    continue

                except NoModelReconstructedError as error:
                    continue

    def _init_loggers(self, request):
        log_dir = create_log_dir()
        iteration_log_file = join(log_dir, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")
        iteration_logger.addHandler(iteration_handler)

        request.data_source.logger = iteration_logger
        request.constructor.logger = iteration_logger
        request.feature_selector.logger = iteration_logger
        request.reconstructor.logger = iteration_logger
        request.deconstructor.logger = iteration_logger
        request.deconstructor.knowledge_database.inject_logger(iteration_logger)

        if request.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)

        return (request.data_source,
                request.constructor,
                request.feature_selector,
                request.reconstructor,
                request.deconstructor)


def create_log_dir():
    log_dir = join(expanduser('~'),
                   ".cml",
                   datetime.now().strftime("%d%m%Y%H%M%S"))
    mkdir(log_dir)
    return log_dir

